import 'dart:collection';

import 'package:flutter/foundation.dart';
import 'package:flutter_mediator/mediator/controller.dart';

import 'assert.dart';
import 'rx/rx_impl.dart';
import 'subscriber.dart';

class Pub {
  Pub()
      : publish = dummyCallback,
        assert(() {
          if (stateModels[runtimeType] != null)
            throw FlutterError(
                'Duplicate model type registered to the `Host` of <$runtimeType>');
          return true;
        }()) {
    ///
    /// class initial steps
    /// 1. rx variables initial
    /// 2. model extends from Pub and initial
    /// 3. set the Pub of rx variables with this
    RxImpl.setPub(this);
    print('RxImpl set pub to: $this');

    stateModels[runtimeType] = this;
    init();
  }

  @mustCallSuper
  @protected
  void init() {}

  //* Map of the models section:
  /// Dependency injection of models, map of models
  static final stateModels = HashMap<Type, Pub>();

  //* static: Get the model.
  static Model getModel<Model extends Pub>() {
    assert(ifStateModel<Model>(stateModels[Model]));
    return stateModels[Model] as Model;
  }
  //! end section

  //* publish aspect section:
  /// Model may update variables before Host init state
  static void dummyCallback([Object aspects]) {}

  /// The publish function, the setState of the host state.
  PublishFn publish;

  /// For reference of regAspects of the host state.
  HashSet<Object> _regAspects;
  HashSet<Object> get regAspects => _regAspects;

  /// For reference of frameAspects of the host state.
  HashSet<Object> _frameAspects;
  HashSet<Object> get frameAspects => _frameAspects;

  /// Set callback and aspects information of the model.
  void setCallback(
    PublishFn publish,
    HashSet<Object> regAspects,
    HashSet<Object> frameAspects,
  ) {
    this.publish = publish;
    _regAspects = regAspects;
    _frameAspects = frameAspects;
  }

  /// Clear the callback and aspects information.
  void restoreCallback() {
    publish = dummyCallback;
    _regAspects = null;
    _frameAspects = null;
  }
  //! end section

  //* subMap section:
  /// Map of subscriber create functions of the model.
  final subMap = HashMap<Object, Object>();

  //* static: Create a `Subscriber<Model>` widget from the `subMap` by the [mapKey] of the `<Model>`.
  static Subscriber<Model> sub<Model extends Pub>(Object mapKey, {Key key}) {
    assert(ifModelTypeCorrect(Model, 'Pub.sub'));

    final model = getModel<Model>();
    final sub = model._getSub<Model>(mapKey);
    return rxSub<Model>(sub, key: key);
  }

  /// Add a create function to the subMap with the key:[o].
  void regSub<Model extends Pub>(Object key, CreatorFn<Model> sub) {
    assert(ifModelTypeCorrect(Model, 'regSub'));
    assert(shouldNull(subMap[key], 'regSub: Duplicate key, subMap[$key]'));

    subMap[key] = sub;
  }

  /// Get the create function by the [key].
  CreatorFn<Model> _getSub<Model extends Pub>(Object key) {
    assert(ifModelTypeCorrect(Model, 'getSub'));
    assert(shouldExists(subMap[key], 'getSub: subMap[$key] not exist'));
    assert(subMap[key] is CreatorFn<Model>);

    return subMap[key] as CreatorFn<Model>;
  }
  //! end subMap section

  //* conMap section:
  /// Map of controller create functions of the model.
  final conMap = HashMap<Object, Object>();

  //* static: Create a `Controller<Model>` widget from the `conMap` by the [mapKey] of the `<Model>`.
  static Controller<Model> con<Model extends Pub>(Object mapKey, {Key key}) {
    assert(ifModelTypeCorrect(Model, 'Pub.con'));

    final model = getModel<Model>();
    final con = model._getCon<Model>(mapKey);
    return Controller<Model>(
      create: con,
      key: key,
    );
  }

  /// Add a create function to the conMap with the key:[o].
  void regCon<Model extends Pub>(Object key, CreatorFn<Model> con) {
    assert(ifModelTypeCorrect(Model, 'regCon'));
    assert(shouldNull(conMap[key], 'regCon: Duplicate key, conMap[$key]'));

    conMap[key] = con;
  }

  /// Get the create function by the [key].
  CreatorFn<Model> _getCon<Model extends Pub>(Object key) {
    assert(ifModelTypeCorrect(Model, 'getCon'));
    assert(shouldExists(conMap[key], 'getCon: conMap[$key] not exist'));
    assert(conMap[key] is CreatorFn<Model>);

    return conMap[key] as CreatorFn<Model>;
  }
  //! end subMap section

}

typedef PublishFn = void Function([Object aspects]);
// typedef PublishFn<AspectType> = void Function([AspectType aspects]);
